# -*- coding: utf-8 -*-
"""Perform random walks on sparse csr matrix."""
from libc.stdlib cimport rand, RAND_MAX

import cython
cimport numpy as np
import numpy as np


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef Py_ssize_t _choose_one(double [:] pmf) nogil:
    """Random choice with discrete probabilities.

    Args:
        pmf (double[:]): probability mass function
    """
    cdef:
        Py_ssize_t i, length
        double random, total

    random = rand() / (RAND_MAX + 1.0)
    length = pmf.shape[0]
    i = 0
    total = 0.0

    while total < random and i < length:
        total += pmf[i]
        i += 1
    return i - 1


@cython.profile(False)
@cython.boundscheck(False)
@cython.wraparound(False)
def walk_random(normalized_csr, np.ndarray labels, int walk_length):
    """Generate random walks for each node in a normalized sparse csr matrix.

    Args:
        normalized_csr (scipy.sparse.csr_matrix): normalized adjacency matrix
        labels (np.ndarray): array of node labels
        walk_length (int): length of walk

    Returns:
        np.array walks, np.array word frequencies
    """
    cdef:
        int [:] indices = normalized_csr.indices
        int [:] indptr = normalized_csr.indptr
        double [:] data = normalized_csr.data
        int num_nodes = normalized_csr.shape[0]
        long [:] vocab_cnt = np.ones(num_nodes, dtype=int)

        int [:] neighbors
        double [:] weights
        int i, j, node_index, weight_index, next_index

    walks = np.empty([num_nodes, walk_length], dtype=object)
    walks.fill('')
    walks.T[0] = labels

    for i in range(num_nodes):
        node_index = i
        for j in range(walk_length-1):
            neighbors = indices[indptr[node_index]:indptr[node_index+1]]
            if neighbors.shape[0] == 0:  # stop walk
                break
            weights = data[indptr[node_index]:indptr[node_index+1]]
            weight_index = _choose_one(weights)
            next_index = neighbors[weight_index]
            walks[i][j+1] = labels[next_index]
            vocab_cnt[next_index] += 1
            node_index = next_index

    return walks, np.asarray(vocab_cnt)



# -*- coding: utf-8 -*-
"""py.test unittests"""
import os
import tempfile

try:
    from unittest import mock
except ImportError:
    import mock

import gensim
import numpy as np
import scipy.sparse as sps

from jwalk import corpus
from jwalk import graph
from jwalk import io
from jwalk import skipgram
from jwalk import __main__

DIR_PATH = os.path.dirname(os.path.realpath(__file__))
KARATE_EDGELIST = os.path.join(DIR_PATH, 'data/karate.edgelist')
KARATE_EMBEDDINGS = os.path.join(DIR_PATH, 'data/karate.embeddings')
KARATE_GRAPH = os.path.join(DIR_PATH, 'data/karate.npz')
TEST_CORPUS = os.path.join(DIR_PATH, 'data/corpus.txt.gzip')

TEST_LABELS = np.array(['A', 'B', 'C'])
TEST_CSR = sps.csr_matrix([[0.0, 1.0, 0.0],
                           [0.0, 0.0, 0.0],
                           [1.0, 0.0, 0.0]])


def test_normalize_csr_matrix():
    normalized = corpus.normalize_csr_matrix(TEST_CSR)
    assert np.array_equal(normalized.todense(), TEST_CSR.todense())


def test_make_undirected():
    undirected = graph.make_undirected(TEST_CSR)
    assert np.array_equal(undirected.todense(), [[0.0, 1.0, 1.0],
                                                 [1.0, 0.0, 0.0],
                                                 [1.0, 0.0, 0.0]])


def test_random_walks():
    normalized = corpus.normalize_csr_matrix(TEST_CSR)
    random_walks, vocab_cnt = corpus.walk_random(normalized, TEST_LABELS,
                                                 walk_length=3)
    assert np.array_equal(random_walks, [['A', 'B', ''],
                                         ['B', '', ''],
                                         ['C', 'A', 'B']])
    assert np.array_equal(vocab_cnt, [2, 3, 1])


def test_walk_graph():
    random_walks, word_freq = corpus.walk_graph(TEST_CSR, TEST_LABELS,
                                                walk_length=3, num_walks=2)
    assert np.array_equal(random_walks, [['A', 'B', ''],
                                         ['B', '', ''],
                                         ['C', 'A', 'B'],
                                         ['A', 'B', ''],
                                         ['B', '', ''],
                                         ['C', 'A', 'B']])
    assert word_freq == {'A': 4, 'B': 6, 'C': 2}


def test_encode_edges():
    edges = np.array([['A', 'B'],
                      ['A', 'C'],
                      ['B', 'C'],
                      ['B', 'E']])
    nodes = np.unique(edges)
    encoded = graph.encode_edges(edges, nodes)
    assert np.array_equal(encoded, [[0, 1],
                                    [0, 2],
                                    [1, 2],
                                    [1, 3]])


def test_build_adjacency_matrix():
    edges = np.array([['A', 'B'],
                      ['A', 'C'],
                      ['B', 'C'],
                      ['B', 'E']])
    csr_matrix, id2item = graph.build_adjacency_matrix(edges)
    assert len(id2item) == 4
    assert np.array_equal(csr_matrix.todense(), [[0., 1., 1., 0.],
                                                 [0., 0., 1., 1.],
                                                 [0., 0., 0., 0.],
                                                 [0., 0., 0., 0.]])


def test_load_edges():
    edges = io.load_edges(KARATE_EDGELIST, delimiter=' ', has_header=False)
    assert np.array_equal(edges[0], ['1', '32'])
    assert edges.shape == (78, 2)


@mock.patch('jwalk.io.PANDAS_INSTALLED', False)
def test_load_edges_no_pandas():
    edges = io.load_edges(KARATE_EDGELIST, delimiter=' ')
    assert np.array_equal(edges[0], ['1', '32'])
    assert edges.shape == (78, 2)


def test_build_corpus():
    with tempfile.NamedTemporaryFile() as f:
        random_walks, word_freqs = corpus.walk_graph(TEST_CSR, TEST_LABELS)
        corpus_path = corpus.build_corpus(random_walks, outpath=f.name)
        assert corpus_path == f.name


def test_train_model():
    model = skipgram.train_model(TEST_CORPUS, size=50, window=5)
    assert len(model.wv.vocab) == 31
    assert model.window == 5
    assert model.vector_size == 50
    assert model.sg == 1


def test_train_skipgram():
    walk_length = 4
    num_walks = 2
    corpus_count = num_walks * len(TEST_LABELS)
    random_walks, word_freq = corpus.walk_graph(TEST_CSR, TEST_LABELS,
                                                walk_length, num_walks)
    with tempfile.NamedTemporaryFile() as f:
        corpus.build_corpus(random_walks, outpath=f.name)
        model = (skipgram
                 .train_model(f.name, size=50, window=5, word_freq=word_freq,
                              corpus_count=corpus_count))
        assert len(model.wv.vocab) == 3
        assert model.window == 5
        assert model.vector_size == 50


def test_jwalk():
    with tempfile.NamedTemporaryFile() as f:
        res = __main__.jwalk(KARATE_EDGELIST, outfile=f.name, delimiter=' ')
        assert res == f.name


def test_online():
    with tempfile.NamedTemporaryFile() as f:
        res = __main__.jwalk(KARATE_EDGELIST, outfile=f.name, delimiter=' ',
                             model_path=KARATE_EMBEDDINGS)
        assert res == f.name


def test_gensim_load():
    with tempfile.NamedTemporaryFile() as f:
        __main__.jwalk(KARATE_EDGELIST, outfile=f.name, delimiter=' ')
        model = gensim.models.Word2Vec.load(f.name)
        assert '1' in model.wv.vocab


def test_adjacency():
    with tempfile.NamedTemporaryFile() as f:
        res = __main__.jwalk(KARATE_GRAPH, outfile=f.name, delimiter=' ')
        assert res == f.name


def test_parser():
    parser = __main__.create_parser()
    args = parser.parse_args(['--input', 'infile', '--output', 'outfile'])
    assert args.infile == 'infile'
    assert args.outfile == 'outfile'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""jwalk CLI.

Prompt parameters:
  debug:            drop a debugger if an exception is raised
  delimiter:        delimiter for input file
  embedding-size:   dimension of word2vec embedding (default=200)
  has-header:       boolean if csv has header row
  help (-h):        argparse help
  input (-i):       file input (edgelist of 2/3 cols or adjacency matrix)
  log-level (-l)    logging level (default=INFO)
  model (-m):       use a pre-existing model
  num-walks (-n):   number of of random walks per graph (default=1)
  output (-o):      file output
  stats:            boolean to calculate walk statistics [requires pandas]
  undirected:       make graph undirected
  walk-length:      length of random walks (default=10)
  window-size:      word2vec window size (default=5)
  workers:          number of workers (default=multiprocessing.cpu_count)

Notes:
  To load graph as input, file must be of type npz and with keys:
  'data', 'indices', 'indptr', 'shape', 'labels'
  Labels must be labels of the indices.

Usage:
  jwalk -i tests/data/karate.edgelist -o karate.embeddings --delimiter=' '
"""
import sys
import os.path
import logging
import tempfile
import multiprocessing
from argparse import RawDescriptionHelpFormatter, ArgumentParser

from jwalk import (build_adjacency_matrix, build_corpus, train_model,
                   walk_graph, load_edges, load_graph, save_graph)

DIR_PATH = os.path.dirname(os.path.realpath(__file__))

logger = logging.getLogger(__name__)
LOGFORMAT = '%(asctime)s %(name)-12s %(levelname)-8s %(message)s'


def debug_hook(type_, value, tb):
    if hasattr(sys, 'ps1') or not sys.stderr.isatty():
        sys.__excepthook__(type_, value, tb)
    else:
        import traceback
        import pdb
        traceback.print_exception(type_, value, tb)
        print(u"\n")
        pdb.pm()


def create_parser():
    parser = ArgumentParser(description=__doc__,
                            formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument('--debug', action='store_true')
    parser.add_argument('--delimiter')
    parser.add_argument('--embedding-size', default=200, type=int)
    parser.add_argument('--graph-path')
    parser.add_argument('--has-header', action='store_true')
    parser.add_argument('--input', '-i', dest='infile', required=True)
    parser.add_argument('--log-level', '-l', type=str.upper, default='INFO')
    parser.add_argument('--num-walks', default=1, type=int)
    parser.add_argument('--model', '-m', dest='model_path')
    parser.add_argument('--output', '-o', dest='outfile', required=True)
    parser.add_argument('--stats', action='store_true')
    parser.add_argument('--undirected', action='store_true')
    parser.add_argument('--walk-length', default=10, type=int)
    parser.add_argument('--window-size', default=5, type=int)
    parser.add_argument('--workers', default=multiprocessing.cpu_count(),
                        type=int)
    return parser


def main():
    parser = create_parser()
    args = parser.parse_args()
    print("Args: ", args)

    if args.debug:
        sys.excepthook = debug_hook

    numeric_level = getattr(logging, args.log_level, None)
    logging.basicConfig(format=LOGFORMAT)
    logger.setLevel(numeric_level)

    return jwalk(**vars(args))


def jwalk(infile, outfile, num_walks=2, embedding_size=100, window_size=5,
          walk_length=10, delimiter=None, model_path=None, stats=False,
          has_header=False, workers=3, undirected=False, **kw):

    outpath = os.path.join(DIR_PATH, '../output')
    if not os.path.exists(outpath):
        os.makedirs(outpath)

    if infile.lower().endswith('.npz'):  # load graph file instead of edges
        logger.debug("Detected npz extension. Assuming input is CSR matrix.")
        logger.info("Loading graph from %s", infile)
        graph, labels = load_graph(infile)
    else:
        logger.info("Loading edges from %s", infile)
        edges = load_edges(infile, delimiter, has_header)
        logger.debug("Loaded edges of shape %s", edges.shape)

        logger.info("Building adjacency matrix")
        graph, labels = build_adjacency_matrix(edges, undirected)
        logger.debug("Number of unique nodes: %d", len(labels))

        graph_path = os.path.join(outpath, 'graph.npz')
        logger.info("Saving graph to %s", graph_path)
        save_graph(graph_path, graph, labels)

    logger.info("Doing %d random walks of length %d", num_walks, walk_length)
    random_walks, word_freq = walk_graph(graph, labels, walk_length, num_walks,
                                         workers)
    logger.debug("Walks shape: %s", random_walks.shape)

    if stats:
        import pandas as pd
        df = pd.DataFrame(random_walks)
        unique_nodes_in_path = df.apply(lambda x: x.nunique(), axis=1)
        logger.info("Unique nodes per walk description: \n" +
                    unique_nodes_in_path.describe().__repr__())

    logger.info("Building corpus from walks")
    with tempfile.NamedTemporaryFile(delete=False) as f_corpus:
        build_corpus(random_walks, outpath=f_corpus.name)

        logger.info("Running Word2Vec on corpus")
        corpus_count = len(labels) * num_walks
        model = train_model(f_corpus.name, embedding_size, window_size,
                            workers=workers, model_path=model_path,
                            word_freq=word_freq, corpus_count=corpus_count)
        model.save(outfile)
        logger.info("Model saved: %s", outfile)

    return outfile



# -*- coding: utf-8 -*-
"""Build encoded sparse csr matrix."""
import logging

import numpy as np
import scipy.sparse as sps

__all__ = ['build_adjacency_matrix', 'encode_edges']

logger = logging.getLogger(__name__)


def encode_edges(edges, nodes):
    """Encode data with dictionary

    Args:
        edges (np.ndarray): np array of the form [node1, node2].
        nodes (np.array): list of unique nodes

    Returns:
        np.ndarray: relabeled edges

    Examples:
        >>> import numpy as np
        >>> edges = np.array([['A', 'B'], ['A', 'C']])
        >>> nodes = np.array(['C', 'B', 'A'])
        >>> print(encode_edges(edges, nodes))
        [[2 1]
         [2 0]]
    """
    sidx = nodes.argsort()
    relabeled_edges = sidx[np.searchsorted(nodes, edges, sorter=sidx)]
    return relabeled_edges


def build_adjacency_matrix(edges, undirected=False):
    """Build adjacency matrix.

    Args:
        edges (np.ndarray): a 2 or 3 dim array of the form [src, tgt, [weight]]
        undirected (bool): if True, add matrix with its transpose

    Returns:
        scipy.sparse.csr_matrix: adjacency matrix, np.ndarray: labels
    """
    assert edges.shape[1] in [2, 3], "Input must contain 2 or 3 columns"

    if edges.shape[1] == 2:  # if no weights
        logger.info("Weight column not found. Defaulting to value 1.")
        weights = np.ones(edges.shape[0], dtype='float')
    else:
        weights = edges[:, 2].astype('float')

    edges = edges[:, :2]
    nodes = np.unique(edges)  # returns sorted
    num_nodes = nodes.shape[0]

    encoded = encode_edges(edges, nodes)

    sp = sps.csr_matrix((weights, encoded.T), shape=(num_nodes, num_nodes))

    if undirected:
        sp = make_undirected(sp)
    return sp, nodes


def make_undirected(csr_matrix):
    """Make CSR matrix undirected."""
    return csr_matrix + csr_matrix.T



# -*- coding: utf-8 -*-
"""Load and save data."""
import logging

import numpy as np
import scipy.sparse as sps

try:
    import pandas as pd
except ImportError:
    PANDAS_INSTALLED = False
else:
    PANDAS_INSTALLED = True

__all__ = ['load_edges', 'load_graph', 'save_graph']

logger = logging.getLogger(__name__)


def load_edges(fpath, delimiter=None, has_header=False):
    """Load edges in CSV format as numpy ndarray of strings.

    Args:
        fpath (str): edges file
        delimiter (str): alternative argument name for sep (default=None)
        has_header (bool): True if has header row

    Returns:
        np.ndarray: array of edges
    """
    if PANDAS_INSTALLED:
        header = 'infer' if has_header else None
        df = pd.read_csv(fpath, delimiter=delimiter, header=header)
        edges = df.values
    else:
        logger.warning("Pandas not installed. Using numpy to load csv, which "
                       "is slower.")
        header = 1 if has_header else 0
        edges = np.genfromtxt(fpath, delimiter=delimiter, skip_header=header,
                              dtype=object)
    return edges.astype('str')


def save_graph(filename, csr_matrix, labels=None):
    np.savez(filename,
             data=csr_matrix.data,
             indices=csr_matrix.indices,
             indptr=csr_matrix.indptr,
             shape=csr_matrix.shape,
             labels=labels)
    return filename


def load_graph(filename):
    loader = np.load(filename)
    sp = sps.csr_matrix((loader['data'], loader['indices'], loader['indptr']),
                        shape=loader['shape'])
    return sp, loader['labels']



# -*- coding: utf-8 -*-
# flake8: noqa
"""jwalk library.

:copyright: (c) 2017 by JW Player.
:license: Apache 2.0, see LICENSE for more details.
"""
import logging

__title__ = 'jwalk'
__author__ = 'Kamil Sindi, Nir Yungster'
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright 2017 JW Player'
# __version__ = get_version(root='..', relative_to=__file__)

# Set default logging handler to avoid "No handler found" warnings.
try:  # Python 2.7+
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

logging.getLogger(__name__).addHandler(NullHandler())

__all__ = []
from .graph import *

__all__ += graph.__all__

from .corpus import *

__all__ += corpus.__all__

from .skipgram import *

__all__ += skipgram.__all__

from .io import *

__all__ += io.__all__



# -*- coding: utf-8 -*-
"""Build word2vec model."""
import os
import logging

from gensim.models import Word2Vec
from gensim.models.word2vec import LineSentence

__all__ = ['train_model']

logger = logging.getLogger(__name__)


def train_model(corpus, size=200, window=5, workers=3, model_path=None,
                word_freq=None, corpus_count=None):
    """Train using Skipgram model.

    Args:
        corpus (str):       file path of corpus
        size (int):         embedding size (default=200)
        window (int):       window size (default=5)
        workers (int):      number of workers (default=3)
        model_path (str):   file path of model we want to update
        word_freq (dict):   dictionary of word frequencies
        corpus_count (int): corpus size

    Returns:
        Word2Vec: word2vec model
    """
    sentences = LineSentence(corpus)
    if model_path is not None:
        logger.info("Updating pre-existing model: %s", model_path)
        assert os.path.isfile(model_path), "File does not exist"
        model = Word2Vec.load(model_path)
        model.build_vocab(sentences, update=True)
        model.train(sentences, total_examples=model.corpus_count,
                    epochs=model.iter)
    else:
        model = Skipgram(sentences=sentences, size=size, window=window,
                         min_count=1, workers=workers, raw_vocab=word_freq,
                         corpus_count=corpus_count)
    return model


class Skipgram(Word2Vec):
    """A subclass to allow more customization of the Word2Vec internals."""

    def __init__(self, raw_vocab=None, corpus_count=None, **kwargs):
        self.raw_vocab = raw_vocab
        self.corpus_count = corpus_count
        super(Skipgram, self).__init__(sg=1, **kwargs)

    def scan_vocab(self, sentences, progress_per=10000, trim_rule=None):
        """Override to supply own word frequencies."""
        if self.corpus_count is None or self.raw_vocab is None:
            super(Skipgram, self).scan_vocab(sentences, progress_per, trim_rule)  # NOQA: E501



# -*- coding: utf-8 -*-
"""Generate text corpus from random walks on graph."""
import numpy as np
from joblib import Parallel, delayed
from joblib.pool import has_shareable_memory

from jwalk import walks

__all__ = ['walk_graph', 'build_corpus']


def walk_random(normalized_csr, labels, walk_length):
    """Generate random walks for each node in a normalized sparse csr matrix.

    Args:
        normalized_csr (scipy.sparse.csr_matrix): normalized adjacency matrix
        labels (np.ndarray): array of node labels
        walk_length (int): length of walk

    Returns:
        np.array walks, np.array word frequencies
    """
    # need to wrap walks.walk_random otherwise joblib complains in Py2
    return walks.walk_random(normalized_csr, labels, walk_length)


def normalize_csr_matrix(csr_matrix):
    """Normalize adjacency matrix weights.

    Args:
        scipy.sparse.csr_matrix: adjacency matrix

    Returns:
        scipy.sparse.csr_matrix
    """
    row_sums = np.array(csr_matrix.sum(axis=1))[:, 0]
    row_indices, col_indices = csr_matrix.nonzero()

    normalized = csr_matrix.copy()
    normalized.data /= row_sums[row_indices]
    return normalized


def walk_graph(csr_matrix, labels, walk_length=40, num_walks=1, n_jobs=1):
    """Perform random walks on adjacency matrix.

    Args:
        csr_matrix: adjacency matrix.
        labels: list of node labels where index align with CSR matrix
        walk_length: maximum length of random walk (default=40)
        num_walks: number of walks to do for each node
        n_jobs: number of cores to use (default=1)

    Returns:
        np.ndarray: list of random walks
    """
    normalized = normalize_csr_matrix(csr_matrix)

    results = (Parallel(n_jobs=n_jobs)
               (delayed(walk_random, has_shareable_memory)
                (normalized, labels, walk_length)
                for _ in range(num_walks)))

    walks, freqs = zip(*results)

    random_walks = np.concatenate(walks)
    word_freqs = np.sum(freqs, axis=0)

    return random_walks, dict(zip(labels, word_freqs))


def build_corpus(walks, outpath):
    """Build corpus by shuffling and then saving as text file.

    Args:
        walks: random walks
        outpath: file to write to

    Returns:
        str: file path of corpus
    """
    np.random.shuffle(walks)
    np.savetxt(outpath, walks, delimiter=' ', fmt='%s')
    return outpath



# -*- coding: utf-8 -*-
"""Distutils setup file, used to install or test 'jwalk'."""
from __future__ import print_function

import sys
import textwrap
import pkg_resources
from setuptools import setup, find_packages, Extension

with open('README.rst') as f:
    readme = f.read()


def is_installed(requirement):
    try:
        pkg_resources.require(requirement)
    except pkg_resources.ResolutionError:
        return False
    else:
        return True


if not is_installed('numpy'):
    print(textwrap.dedent("""
            Error: numpy needs to be installed first. You can install it via:

            $ pip install numpy
            """), file=sys.stderr)
    exit(1)


def ext_modules():
    import numpy

    walks_ext = Extension('jwalk.walks', ['jwalk/src/walks.pyx'],
                          include_dirs=[numpy.get_include()])

    return [walks_ext]


setup(
    name='jwalk',
    description='Representational learning on graphs',
    long_description=readme,
    packages=find_packages(exclude=['tests', 'docs']),
    use_scm_version=True,
    ext_modules=ext_modules(),
    author='Kamil Sindi, Nir Yungster',
    author_email='kamil@jwplayer.com, nir@jwplayer.com',
    url='https://github.com/jwplayer/jwalk',
    install_requires=[
        'setuptools_scm>=1.15.0',
        'Cython',
        'numpy',
        'scipy',
        'gensim>=2.0.0',
        'joblib',
    ],
    setup_requires=[
        'setuptools>=18.0',
        'Cython>=0.20',
        'numpy',
        'pytest-runner',
        'setuptools_scm>=1.15.0',
        'sphinx_rtd_theme',
    ],
    tests_require=[
        'pytest',
        'pytest-flake8',
    ],
    extras_require={
        'all': ['pandas'],
        ':python_version=="2.7"': ['mock'],
    },
    zip_safe=False,
    include_package_data=True,
    classifiers=textwrap.dedent("""
        Development Status :: 4 - Beta
        Intended Audience :: Science/Research
        License :: OSI Approved :: Apache Software License
        Environment :: Console
        Programming Language :: Python :: 2
        Programming Language :: Python :: 2.7
        Programming Language :: Python :: 3
        Programming Language :: Python :: 3.4
        Programming Language :: Python :: 3.5
        Programming Language :: Python :: 3.6
        Programming Language :: Cython
        Topic :: Scientific/Engineering
        """).strip().splitlines(),
    keywords=['deep learning', 'neural networks', 'deepwalk'],
    license='Apache License 2.0',
    entry_points={
        'console_scripts': [
            'jwalk=jwalk.__main__:main'
        ]
    },
)
